#nanite_fight: instructions.py
#By Shanti Pothapragada, rgbDreamer@gmail.com
#for pyweek 12: 'nine times'
#under GPL 3.0

import random

import pygame

import view_controller

randint = random.Random().randint

#cardinal directions:
# 0
#3+1
# 2
dirs = {0:(0,-1), 1:(1,0), 2:(0,1), 3:(-1,0)}
bullet_time = 400#ms
bullet_freeze = 3750
shield_time = 2000#ms
shield_freeze = 2500


def forward(n, game):
	dir = dirs[n[2]]
	n[0] = (n[0] + dir[0]) % 9
	n[1] = (n[1] + dir[1]) % 9
	
def left(n, game):
	dir = dirs[(n[2] - 1) % 4]
	n[0] = (n[0] + dir[0]) % 9
	n[1] = (n[1] + dir[1]) % 9
	
def right(n, game):
	dir = dirs[(n[2] + 1) % 4]
	n[0] = (n[0] + dir[0]) % 9
	n[1] = (n[1] + dir[1]) % 9
	
def back(n, game):
	dir = dirs[(n[2] + 2) % 4]
	n[0] = (n[0] + dir[0]) % 9
	n[1] = (n[1] + dir[1]) % 9
	
def wait(n, game):
	pass
	
def shoot(n, game):
	dir = dirs[n[2]]
	n[6] += bullet_freeze
	#x, y, dir ((-1,0) style), spaces left, timer, team
	
	return [(n[0] + dir[0]) % 9, (n[1] + dir[1]) % 9, dir, 3, bullet_time, n[3]]
	
def shield(n, game):
	n[6] += shield_freeze
	return [n[0], n[1], shield_time]
	
def turnCW(n, game):
	n[2] = (n[2] + 1) % 4
	
def turnCCW(n, game):
	n[2] = (n[2] - 1) % 4
	
actions = [turnCCW, left, forward, back, right, turnCW, shoot, wait, shield]
	
def neutral_random(n, game):
	action = actions[randint(0, 6)]
	n[6] += costs[action]
	action(n, game)
	view_controller.instruction(action, n)
	
	
def goto(n, function):
	n[4] = function - 1 #index from 0
	n[5] = -1 #so it gets incremented to 0
	
def makeGoto(function):
	return lambda n, game: goto(n, function)	
	
calls = [makeGoto(1), makeGoto(2), makeGoto(3), makeGoto(4), 
		makeGoto(5), makeGoto(6), makeGoto(7), makeGoto(8), makeGoto(9)]
		
def condition_report(n, bool):
	if n[8] < 0:
		bool = not bool	# not
		n[8] = - n[8]
	if n[8] == 1: n[7] = bool	# replace
	if n[8] == 2: n[7] = n[7] and bool	# and
	if n[8] == 3: n[7] = n[7] or  bool	# or
	n[8] = 1	#reset to replace mode.
	
def if_condition(n, game):
	if not n[7]:	#if flag is False,
		n[5] += 1	#skip next instruction.
	
def or_condition(n, game):
	n[8] = 3
	
def and_condition(n, game):
	n[8] = 2
	
def not_condition(n, game):
	n[8] = -n[8]
		
def is_front(n, game):
	nanites = game[0]
	x, y = n[0], n[1]
	fx, fy = dirs[n[2]]				#forward
	tiles = [
		(x + fx * 1, y + fy * 1),		#1
		(x + fx * 2, y + fy * 2),		#2
		(x + fx * 3, y + fy * 3)]		#3
	for other in nanites:
		for tile in tiles:
			if other[0] == tile[0] and other[1] == tile[1]:
				condition_report(n, True)
				return
	condition_report(n, False)
	
def is_left(n, game):
	nanites = game[0]
	#  1 
	# 23 
	#456^     ^ = ship pointed up
	# 789
	#  AB
	#   C
	x, y = n[0], n[1]
	fx, fy = dirs[n[2]]				#forward
	lx, ly = dirs[(n[2] - 1) % 4] 	#left
	tiles = [
		(x + fx * 2 + lx * 1, y + fy * 2 + ly * 1),		#1
		(x + fx * 1 + lx * 1, y + fy * 1 + ly * 1),		#2
		(x + fx * 1 + lx * 2, y + fy * 1 + ly * 2),		#3
		(x + fx * 0 + lx * 3, y + fy * 0 + ly * 3),		#4
		(x + fx * 0 + lx * 2, y + fy * 0 + ly * 2),		#5
		(x + fx * 0 + lx * 1, y + fy * 0 + ly * 1),		#6
		(x + fx *-1 + lx * 2, y + fy *-1 + ly * 2),		#7
		(x + fx *-1 + lx * 1, y + fy *-1 + ly * 1),		#8
		(x + fx *-1 + lx * 0, y + fy *-1 + ly * 0),		#9
		(x + fx *-2 + lx * 2, y + fy *-2 + ly * 2),		#A
		(x + fx *-2 + lx * 1, y + fy *-2 + ly * 1),		#B
		(x + fx *-3 + lx * 0, y + fy *-3 + ly * 0)]		#C
	for other in nanites:
		for tile in tiles:
			if other[0] == tile[0] and other[1] == tile[1]:
				condition_report(n, True)
				return
	condition_report(n, False)
	
def is_right(n, game):
	nanites = game[0]
	#the mirror of isleft.  They double-cover straight behind.
	x, y = n[0], n[1]
	fx, fy = dirs[n[2]]				#forward
	lx, ly = dirs[(n[2] + 1) % 4] 	#right
	tiles = [
		(x + fx * 2 + lx * 1, y + fy * 2 + ly * 1),		#1
		(x + fx * 1 + lx * 1, y + fy * 1 + ly * 1),		#2
		(x + fx * 1 + lx * 2, y + fy * 1 + ly * 2),		#3
		(x + fx * 0 + lx * 3, y + fy * 0 + ly * 3),		#4
		(x + fx * 0 + lx * 2, y + fy * 0 + ly * 2),		#5
		(x + fx * 0 + lx * 1, y + fy * 0 + ly * 1),		#6
		(x + fx *-1 + lx * 2, y + fy *-1 + ly * 2),		#7
		(x + fx *-1 + lx * 1, y + fy *-1 + ly * 1),		#8
		(x + fx *-1 + lx * 0, y + fy *-1 + ly * 0),		#9
		(x + fx *-2 + lx * 2, y + fy *-2 + ly * 2),		#A
		(x + fx *-2 + lx * 1, y + fy *-2 + ly * 1),		#B
		(x + fx *-3 + lx * 0, y + fy *-3 + ly * 0)]		#C
	for other in nanites:
		for tile in tiles:
			if other[0] == tile[0] and other[1] == tile[1]:
				condition_report(n, True)
				return
	condition_report(n, False)

def is_bullet(n, game):	
	for i in [-3,-2,-1,1,2,3]:
		for bullet in game[1]:
			if bullet[0] == n[0] + i and bullet[1] == n[1]:
				condition_report(n, True)
				return	
	for j in [-3,-2,-1,1,2,3]:
		for bullet in game[1]:
			if bullet[0] == n[0] and bullet[1] == n[1] + j:
				condition_report(n, True)
				return
	condition_report(n, False)
	
def is_near(n, game):	
	nanites = game[0]
	for i in [-2,-1,1,2]:
		for other in nanites:
			if other[0] == n[0] + i and other[1] == n[1]:
				condition_report(n, True)
				return	
	for j in [-2,-1,1,2]:
		for other in nanites:
			if other[0] == n[0] and other[1] == n[1] + j:
				condition_report(n, True)
				return
	for i in [-1,1]:
		for j in [-1, 1]:
			if other[0] == n[0] + i and other[1] == n[1] + j:
				condition_report(n, True)
				return
	condition_report(n, False)
	
conditions = [if_condition, not_condition, or_condition, and_condition,
	is_front, is_near, is_left, is_right, is_bullet]
costs = {turnCCW:500, left:1000, right:800, forward:750, back:800, 
		turnCW:500, shoot:250, wait:500, shield:0, neutral_random:0,
		if_condition: 50, not_condition:25, or_condition:25, and_condition:25,
		is_front:250, is_near:250, is_left:250, is_right:250, is_bullet:250}
for call in calls:
	costs[call] = 50
		