#nanite_fight: view_controller.py
#By Shanti Pothapragada, rgbDreamer@gmail.com
#for pyweek 12: 'nine times'
#under GPL 3.0

import sys,os

import pygame
from pygame.mixer import Sound
from pygame.locals import *
from random import Random

from instructions import actions, conditions, calls, wait

#----------------------------------Variable initiaion-------------------------

index = Random().randint(0,4)
screen = pygame.display.set_mode((792,312))
tileSize = 24

volume = 1
player_volume = .9
neutral_volume = .1

arena_offset = 12, 1
arenaTileSize = 24
code_offset = (2, 2), (23, 2)
actions_offset = (2, 1), (23, 1)
calls_offset = (1, 2), (22, 2)
conditions_offset = (2, 11), (23, 11)
settings_offset = (12, 11)

vertical_layout = zip([0,0,0,0,0,0,0,0,0],[0,1,2,3,4,5,6,7,8])
horizontal_layout = zip([0,1,2,3,4,5,6,7,8],[0,0,0,0,0,0,0,0,0])
square_layout = zip([0,1,2,0,1,2,0,1,2], [0,0,0,1,1,1,2,2,2])
function_square_layout = zip([0,3,6,0,3,6,0,3,6],[0,0,0,3,3,3,6,6,6])
function_lines_layout = vertical_layout

_offsets = {'action':actions_offset, 'condition':conditions_offset, 
	'call':calls_offset, 'code':code_offset, 'arena':arena_offset,
	'setting':settings_offset}
layouts = {'action': horizontal_layout, 'condition': horizontal_layout, 
	'call':vertical_layout, 'code': horizontal_layout, 
	'function':function_lines_layout, 'setting':horizontal_layout}

functions = {'action': actions, 'condition':conditions, 'call':calls}
players = [1,2]
playersVisible = [1,2]
	
tileDims = 33, 13
mousepos = 0,290
shift = False
dragging = False
selected = None # this will be a (function, image) pair. 
operations = None 	#holds a reference to nanite_fight.operations, set by
					#set_operations(operations).
tooltip = ['this is a test of the emergency broadcast system.',
'oh no the grey goo',
'welcome to nanite_fight',
'everything is sooo small',
"I'm a tooltip, not a help balloon",][index]
feedback = ['nanite_fight', pygame.time.get_ticks() + 1000]
font = pygame.font.SysFont(name = None, size = 14)
font_feedback = pygame.font.SysFont(name = None, size = 20)
font_big = pygame.font.SysFont(None, 30)
font_height = font.get_height()
#maps used for input:
action_map = {}
settings_map = {}
operations_map = {}




#cardinal directions:
# 0
#3+1
# 2
dirs = {0:90, 1:0, 2:270, 3:180}



#-----------------------------settings functions------------------------------

def load_image(filename, colorkey = (255,255,255), alpha = False):
	if alpha: 
		image = pygame.image.load(filename).convert_alpha()
	else:
		image = pygame.image.load(filename).convert()
	image.set_colorkey(colorkey)
	return image
	
	
def save(plr):
	filename = 'player %s 0.sav'%(plr)
	files = os.listdir('saves')
	i = 2
	while filename in files:
		filename = filename[:9] + str(i) + '.sav'
		i += 1
	filename = 'saves/' + filename
	file = open(filename, 'w')
	data = ''
	string = ''
	for function in operations[plr]:
		functionString = ''
		for instruction in function:
			functionString += instruction_to_int(instruction) + ','
		string += functionString + '\n'
		
	string += str(player_colors[plr-1])
	
	file.write(string)
	file.close()
	feedback[0] = 'Saved Player %s'%(plr)
	feedback[1] = pygame.time.get_ticks() + 1000
	
def load(plr):
	import loader
	loader.main(plr)
	feedback[0] = 'Loaded Player %s'%(plr)
	feedback[1] = pygame.time.get_ticks() + 1000
	gen_tilemaps()
	
def instruction_to_int(instruction):
	if instruction in actions:
		return str(actions.index(instruction))
	if instruction in calls:
		return str(calls.index(instruction) + 9)
	if instruction in conditions:
		return str(conditions.index(instruction) + 18)
	raise IOError
	
def int_to_instruction(i):
	i = int(i)
	inst = []
	inst.extend(actions)
	inst.extend(calls)
	inst.extend(conditions)
	return inst[i]
	
def toggle_volume():
	global volume, neutral_volume, player_volume
	if volume == 1:
		#player vol
		settingImages[3].fill((255,255,255))
		settingImages[3].blit(load_image("img/vol_player.bmp"), (0,0))
		volume = -1
		neutral_volume = 0
		player_volume = .5
	elif volume == -1:
		#mute
		settingImages[3].fill((255,255,255))
		settingImages[3].blit(load_image("img/vol_mute.bmp"), (0,0))
		volume = 0
		neutral_volume = 0
		player_volume = 0
	else:
		#all
		settingImages[3].fill((255,255,255))
		settingImages[3].blit(load_image("img/vol_all.bmp"), (0,0))
		volume = 1
		neutral_volume = .1
		player_volume = .9
	pygame.mixer.Channel(11).set_volume(player_volume)
	pygame.mixer.Channel(10).set_volume(player_volume)
	
def toggle_speed():
	speed = sys.modules["__main__"].speed 
	if 0 < speed < 1:
		settingImages[4].fill((255,255,255))
		settingImages[4].blit(load_image("img/onespeed.bmp"), (0,0))
		sys.modules["__main__"].speed = 1
	elif speed == 1:
		settingImages[4].fill((255,255,255))
		settingImages[4].blit(load_image("img/threespeed.bmp"), (0,0))
		sys.modules["__main__"].speed = 3
	elif speed == 3:
		settingImages[4].fill((255,255,255))
		settingImages[4].blit(load_image("img/pause.bmp"), (0,0))
		sys.modules["__main__"].speed = 0
	else:
		settingImages[4].fill((255,255,255))
		settingImages[4].blit(load_image("img/thirdspeed.bmp"), (0,0))
		sys.modules["__main__"].speed = .333
		
def quit():
	sys.modules["__main__"].running = False
	
def color_swap(image, new_color, old_color = (0,0,0)):
	pa = pygame.PixelArray(image)
	pa.replace(old_color, new_color)

def choose_color(plr, index = -1):
	if index == -1:
		player_colors[plr-1] = (player_colors[plr-1] + 1) % len(colors)
	elif 0 < index < len(colors):
		player_colors[plr-1] = index % len(colors)
	color1 = colors[player_colors[plr-1]]
	color2 = colors[(player_colors[plr-1] + 1) % len(colors)]
	for src, dest in zip(playerImages[0], playerImages[plr]):
		dest.blit(src, (0,0))
		color_swap(dest, color1, (0,0,0))
		color_swap(dest, color2, (255,255,255))
	
def gen_tilemaps():
	"""this creates maps from tile x,y to functions.
	It is called by set_operations() automatically, but should be 
	called anytime the layout (or these functions) changes.
	
	Note: for operations, returns an array and an index so the value cancel
	be updated. This also means that this doesn't need to be called when
	each operations function changes"""
	global toolbar_map, settings_map, operations_map
	#settings
	settings_map = {}
	for i in range(len(settings)):
		settings_map[
			(_offsets['setting'][0] + layouts['setting'][i][0]),
			(_offsets['setting'][1] + layouts['setting'][i][1])
			] = settings[i]
	#actions
	toolbar_map = {}
	for type in ['action', 'condition', 'call']:
		for i in range(len(functions[type])):
			for plr in players:
				toolbar_map[
				((_offsets[type][plr-1][0] + layouts[type][i][0]),
				(_offsets[type][plr-1][1] + layouts[type][i][1])) 
				] =	functions[type][i]
	#operations
	operations_map = {}
	for i in range(9):
		for j in range(9):
			for plr in playersVisible:
				operations_map[
					((j + code_offset[plr-1][0]),
					 (i + code_offset[plr-1][1]))
					 ] = operations[plr][i], j
	
def set_operations(new_operations):
	global operations
	operations = new_operations
	gen_tilemaps()
	
def set_players(new_players):
	global players
	if new_players == [1,2] or  new_players == [1] or new_players == [2]:
		players = new_players

settings = [
	lambda: choose_color(1,-1),
	lambda: save(1),
	lambda: load(1),
	toggle_volume,
	toggle_speed,
	quit,
	lambda: choose_color(2,-1),
	lambda: save(2),
	lambda: load(2)]
	
		
		
#------------------------------load resources--------------------------------
#load sounds:
pickup = Sound('sfx/pickup.ogg')
putdown = Sound('sfx/putdown.ogg')
cancel = Sound('sfx/cancel.ogg')
beat = Sound('sfx/beat.ogg')
segfault = Sound('sfx/segfault.ogg')

#load graphics:
bg = load_image("img/background6.bmp", None)
shieldImage = load_image("img/shield_effect.png", colorkey = None, alpha = True)
playerImages = [[],[], []]

for plr in playerImages:
	plr.extend([
	load_image("img/nanite.bmp", colorkey = (255,0,255)),
	load_image("img/bullet.bmp", colorkey = (255,0,255)),
	load_image('img/color.bmp', colorkey = (255,0,255)),
	load_image('img/save.bmp', colorkey = (255,0,255)),
	load_image('img/load.bmp', colorkey = (255,0,255))
	])
	
naniteImages = playerImages[0][0], playerImages[1][0], playerImages[2][0]
bulletImages = playerImages[0][1], playerImages[1][1], playerImages[2][1]
settingImages =  [playerImages[1][2], playerImages[1][3], playerImages[1][4], 
				load_image('img/vol_all.bmp'), load_image('img/onespeed.bmp'),
				load_image('img/exit.bmp'), 
				playerImages[2][2], playerImages[2][3], playerImages[2][4]]

				
#setup player colors:
colors = [(250,0,0), (0,0,250), (0,200,0), 
		(0,0,0), (0,200,180),  (100,0,200)]
		
player_colors = [4,2] #red/purple, green/blue

		
choose_color(1, 5)
choose_color(2, 1)

#map resources to game elements:


images = {}
sounds = {}

actionImages = [load_image('img/turnCCW.bmp'), load_image('img/left.bmp'), 
				load_image('img/forward.bmp'), load_image('img/back.bmp'), 
				load_image('img/right.bmp'), load_image('img/turnCW.bmp'), 
				load_image('img/shoot.bmp'), load_image('img/wait.bmp'),
				load_image('img/shield.bmp')]
actionSounds = [Sound('sfx/turn.ogg'), Sound('sfx/slide.ogg'), 
				Sound('sfx/forward.ogg'), Sound('sfx/back.ogg'), 
				Sound('sfx/slide.ogg'), Sound('sfx/turn.ogg'), 
				Sound('sfx/shoot.ogg'), None, 
				Sound('sfx/turn.ogg')]
for action,image in zip(actions, actionImages):
	images[action] = image
for action,sound in zip(actions, actionSounds):
	sounds[action] = sound
sounds[None] = segfault
	
callImages = [load_image('img/1.bmp'), load_image('img/2.bmp'), 
				load_image('img/3.bmp'), load_image('img/4.bmp'), 
				load_image('img/5.bmp'), load_image('img/6.bmp'), 
				load_image('img/7.bmp'), load_image('img/8.bmp'), 
				load_image('img/9.bmp'), ]
for call,image in zip(calls, callImages):
	images[call] = image
callSound = Sound('sfx/call.ogg')
for call in calls:
	sounds[call] = callSound
scanSound = Sound('sfx/scan.ogg')
for condition in conditions[3:]: #the scans- not OR, THEN, etc.
	sounds[condition] = scanSound
	
conditionImages = [load_image('img/if.bmp'), load_image('img/not.bmp'),  
				load_image('img/or.bmp'), load_image('img/and.bmp'),  
				load_image('img/is_front.bmp'),  load_image('img/is_near.bmp'),  
				load_image('img/is_left.bmp'),  load_image('img/is_right.bmp'),  
				load_image('img/is_bullet.bmp')]
for condition,image in zip(conditions, conditionImages):
	images[condition] = image
	
for setting, image in zip(settings, settingImages):
	images[setting] = image

#------------------------controller: input handling--------------------------

def handle_events():
	global dragging, selected, mousepos, shift, tooltip
	tooltip_popup(mousepos)
	for e in pygame.event.get():
		if e.type == pygame.QUIT:
			sys.exit()
		if e.type == pygame.MOUSEBUTTONDOWN:
			if e.button == 3: #cancel
				if selected:
					cancel.play()
				dragging = False
				selected = None
			if e.button == 1:
				if selected:
					drop(e.pos) #drop selected
				else:
					click(e.pos) #select new
		if e.type == pygame.MOUSEMOTION:
			mousepos = e.pos
			tooltip = None
			if e.buttons[0]:
				dragging = True
		if e.type == pygame.MOUSEBUTTONUP:
			if dragging and selected and e.button == 1:
				drop(e.pos) # drop dragged
			pass
		if e.type == pygame.locals.USEREVENT + 1:
			pygame.mixer.Channel(11).play(beat)
		if e.type == pygame.KEYUP:
			if e.key == K_LSHIFT or e.key == K_RSHIFT:
				shift = False
		if e.type == pygame.KEYDOWN:
			if e.key == K_LSHIFT or e.key == K_RSHIFT:
				shift = True
				
def click(pos):
	global selected, dragging, operations
	global actions, conditions, calls
	tile = pos[0] / tileSize, pos[1] / tileSize
	if tile in toolbar_map:
		selected = toolbar_map[tile]
		pickup.play()
	if tile in operations_map:
		function, instruction = operations_map[tile]
		selected = function[instruction]
		function[instruction] = wait
		pickup.play()
	if tile in settings_map:
		settings_map[tile]()
		
def drop(pos):
	global selected, dragging, operations
	tile = pos[0] / tileSize, pos[1] / tileSize
	
	if tile in operations_map:
		function, instruction = operations_map[tile]
		function[instruction] = selected
		putdown.play()
	else:
		cancel.play()
	if not shift:
		selected = None
		dragging = False
		
def tooltip_popup(pos):
	global tooltip
	tile = pos[0] / tileSize, pos[1] / tileSize
	if tile in settings_map:
		tooltip = tooltips[settings_map[tile]]
	if tile in toolbar_map:
		tooltip = tooltips[toolbar_map[tile]]
	
	
tooltips = {}
actionTips = ['Turn CCW: 500ms', 'Slide Left: 1000:ms', 'Move Forward: 750ms',
		'Move Back: 1000ms', 'Slide Right: 1000ms', 'Turn CW: 500ms',
		'Launch Virus: 250ms, fires, freezes for 3750ms\n' +
		' virus travels 3 grids, 400ms each grid','Wait: 500ms',
		'Activate Shield: lasts 2000ms, freezes for 2500ms']
callTips = ['Jump to Function 1: 50 ms', 'Jump to Function 2: 50 ms',
			'Jump to Function 3: 50 ms', 'Jump to Function 4: 50 ms',
			'Jump to Function 5: 50 ms', 'Jump to Function 6: 50 ms', 
			'Jump to Function 7: 50 ms', 'Jump to Function 8: 50 ms', 
			'Jump to Function 9: 50 ms', ]
			
conditionTips = [
	"THEN: 50ms\n"+
	"if the last condition processed was false, skip the next instruction.", 
	"NOT: 25 ms\n"+
	" Flips the next condition processed.",
	"AND: 25 ms\n"+
	" State is True only if the last and next conditions processed are both true.",
	"OR: 25 ms\n"+
	" State is True if either the last and next conditions processed were true.",
	"Front Scan: 250ms\n"+
	" Returns true if there is a nanite 1,2, or 3 grids ahead.",
	"Proximity Scan: 250ms\n"+
	" Returns true if there is a nanite within a 1 grid radius.",
	"Left Scan: 250ms\n"+
	" Returns true if there is a nanite within 2 grid to the left\n"+
	" Overlaps with Right Scan in back, does not scan forward",
	"Right Scan: 250ms\n"+
	" Returns true if there is a nanite within 2 grid to the right\n"+
	" Overlaps with Left Scan in back, does not scan forward",
	"Bullet Scan: 250ms\n"+
	" Returns true if there is a nullet in 3 grids directly up, down, left, or right."
	]
	
settingTips = [
	'Change Player 1 color',
	'Save Player 1 instructions',
	'Load Player 1 instructions',
	'Toggle Sound',
	'Toggle Speed',
	'Quit',
	'Change Player 2 color',
	'Save Player 2 instructions',
	'Load Player 2 instructions',
	]

for tip, function in zip(actionTips, actions):
	tooltips[function] = tip
for tip, function in zip(callTips, calls):
	tooltips[function] = tip
for tip, function in zip(conditionTips, conditions):
	tooltips[function] = tip
for tip, function in zip(settingTips, settings):
	tooltips[function] = tip

	
			
			
#---------------------------view: Output handling------------------------	
	
def instruction(instruction, n):
	"""This is called everytime a nanite does an instruction."""
	#play sounds:
	if instruction in sounds and sounds[instruction]:
		if n[3] in players:#nanite on playing team?
			vol = player_volume
		else:
			vol = neutral_volume
		if vol:
			channel = pygame.mixer.Channel(n[9])
			if channel.get_busy():
				channel = pygame.mixer.find_channel(True)
			channel.set_volume(vol)
			channel.play(sounds[instruction])
	if instruction is None:
		feedback[0] = 'nanite segfault, rebooting'
		feedback[1] = 5000 + pygame.time.get_ticks()
		
def draw(game):
	nanites, bullets, shields, operations = game
	
	#draw background:
	screen.blit(bg, (0, 0))
	
	#draw toolbars:
	for type in ['action', 'condition', 'call']:
		for i in range(len(functions[type])):
			for plr in players:
				screen.blit(images[functions[type][i]],
					((_offsets[type][plr-1][0] + layouts[type][i][0]) * tileSize,
					(_offsets[type][plr-1][1] + layouts[type][i][1]) * tileSize))
					
	#draw settings:
	for i in range(len(settings)):
		screen.blit(images[settings[i]],
			((_offsets['setting'][0] + layouts['setting'][i][0]) * tileSize,
			(_offsets['setting'][1] + layouts['setting'][i][1]) * tileSize))
	
					
	#draw instructions:
	for i in range(9):
		for j in range(9):
			for plr in playersVisible:
				screen.blit(images[operations[plr][i][j]], 
					((j + code_offset[plr-1][0]) * tileSize,
					 (i + code_offset[plr-1][1]) * tileSize))
						 
	#draw nanites:
	for n in nanites:
		tmp = pygame.transform.rotate(playerImages[n[3]][0], dirs[n[2]])
		screen.blit(tmp, ((n[0] + arena_offset[0]) * tileSize, 
						  (n[1] + arena_offset[1]) * tileSize))
		if n[3]:
			if n[4] == -1:# nanite segfaulting
				pygame.draw.aaline(screen, colors[player_colors[n[3]-1]], 
						((n[0] + arena_offset[0]) * tileSize + tileSize / 2, 
						 (n[1] + arena_offset[1]) * tileSize + tileSize / 2), 
						(16.5 * tileSize, 10.5 * tileSize))
			else:	
				pygame.draw.aaline(screen, colors[player_colors[n[3]-1]], 
					((n[0] + arena_offset[0]) * tileSize + tileSize / 2, 
					 (n[1] + arena_offset[1]) * tileSize + tileSize / 2), 
					((n[5] + code_offset[n[3] - 1][0]) * tileSize + tileSize / 2,
					(n[4] + code_offset[n[3] - 1][1]) * tileSize + tileSize / 2))
					
	#draw bullets:
	for b in bullets:
		screen.blit(pygame.transform.rotate(
					playerImages[b[5]][1], 90 * b[2][1]),
					((b[0] + arena_offset[0]) * tileSize, 
					 (b[1] + arena_offset[1]) * tileSize))
	#draw shields:
	for s in shields:
		screen.blit(shieldImage, ((s[0] + arena_offset[0]) * tileSize, 
								  (s[1] + arena_offset[1]) * tileSize))
	#draw feedback
	if pygame.time.get_ticks() >= feedback[1]:
		feedback[0] = None
	if feedback[0]:
		width, height = font_feedback.size(feedback[0])
		screen.blit(font_feedback.render(feedback[0], True, (255,255,255)), 
							(16.5 * tileSize - width / 2, 10 * tileSize + 4))
	
	#draw tooltips:
	if not selected and tooltip:
		text = tooltip.split('\n')
		width = 0
		for string in text:
			tmp = font.size(string)[0]
			if tmp > width: 
				width = tmp
		size = width + 4, font_height * len(text) + 1
		pos = [mousepos[0] + 12, mousepos[1] + 12 - font_height * len(text)]
		if pos[0] + size[0] > screen.get_width():
			pos = [pos[0] - size[0] - 24, pos[1]]
		pygame.draw.rect(screen, (255,255,220), Rect(pos, size))
		pygame.draw.rect(screen, (40,0,60), Rect(pos, size), 1)
		for string in text:
			screen.blit(font.render(string, True, (100,0,100)),
					(pos[0] + 2, pos[1] + 1))
			pos[1] += font_height
	#draw mouse dragging:
	if selected:
		screen.blit(images[selected], 
				(mousepos[0] - tileSize / 2, mousepos[1] - tileSize / 2))
		
	pygame.display.flip()

def game_over(winner):
	string = 'Nanite Team %s Wins!'%(winner)
	color1 = colors[player_colors[winner - 1]]
	color2 = colors[(player_colors[winner - 1] + 1) % len(colors)]
	image = pygame.surface.Surface((900,400), SRCALPHA)
	size = font_big.size('player 3 wins!')
	x,y = 0,0
	for i in range(30):
		for j in range(5):
			image.blit(font_big.render(string, True, color1), (x,y))
			x += size[0] * 1.5 #wtf?
			image.blit(font_big.render(string, True, color2), (x,y))
			x += size[0] * 1.5
		x = 0
		y += size[1]
		color1, color2 = color2, color1
	image = pygame.transform.rotate(image, 12)
	screen.blit(image, (-100,-100))
	pygame.display.flip()
	pygame.time.Clock().tick(1)
	
				
	
		
	
	
	
	
	
	
	