#nanite_fight: nanite_fight.py
#By Shanti Pothapragada, rgbDreamer@gmail.com
#for pyweek 12: 'nine times'
#under GPL 3.0

import random

import pygame
from pygame.locals import *

import code
import code.view_controller
import code.instructions

randint = random.Random().randint
clock = pygame.time.Clock()	
dt = 0
start_delay = 1000
speed = 1
max_dt = 84 #12 frames per second.

nanites = [
	#0x, 1y, 2dir, 3team, 4function, 5instruction, 6timer, 7flag, 8flag mode
	[1, 1, randint(0,3), 0, 0, 0, start_delay, 0, 0, 1], #1
	[4, 1, randint(0,3), 0, 0, 0, start_delay, 0, 0, 2], #2
	[7, 1, randint(0,3), 0, 0, 0, start_delay, 0, 0, 3], #3
	[1, 3, 1,			 1, 0, 0, start_delay, 0, 0, 4], #4: player 1
	[4, 4, randint(0,3), 0, 0, 0, start_delay, 0, 0, 5], #5
	[7, 5, 3, 			 2, 0, 0, start_delay, 0, 0, 6], #6: player 2
	[1, 7, randint(0,3), 0, 0, 0, start_delay, 0, 0, 7], #7
	[4, 7, randint(0,3), 0, 0, 0, start_delay, 0, 0, 8], #8
	[7, 7, randint(0,3), 0, 0, 0, start_delay, 0, 0, 9]] #9
	#dir is 0 for up, 1 for right, 2 for down, 3 for left
	#team is 0 for neutral, 1 or 2 for player 1/2
	#function is the row in the operations table
	#instruction is the column
	#timer is the milliseconds left before a new action
	#flag stores conditions true or false
	#flag mode: 1: replace, 2: and, 3: or, -1: not, -2: and not, -3 or not
bullets = []
shields = []
operations = []
game = [nanites, bullets, shields, operations]


for n in range(3):
	operations.append([])
	for function in range(9):
		operations[-1].append([])
		for instruction in range(8):
			operations[-1][-1].append(code.instructions.wait)
		operations[-1][-1].append(code.instructions.calls[function])
		
#neutral loop a random action:
operations[0][0][0] = code.instructions.neutral_random
operations[0][0][1] = code.instructions.wait
operations[0][0][2] = code.instructions.calls[0]
#two demo actions:
operations[1][0][3] = code.instructions.turnCCW
operations[2][0][4] = code.instructions.turnCW
		
code.view_controller.set_operations(operations)
		

running = True
while running:
	#get input:
	code.view_controller.handle_events()
	#update nanites:
	for n in nanites:
		tmp = dt
		while tmp > 0:
			if n[6] > 0:
				n[6], tmp = n[6] - tmp , tmp - n[6] #update timers
			else: #do next op
				if n[5] == -1: #nanite was rebooting.
					n[4], n[5] = 0,0
					continue
				instruction = operations[n[3]][n[4]][n[5]]
				make = instruction(n, game) #call function
				code.view_controller.instruction(instruction, n)
				n[5] += 1
				#wrap functions/instructions:
				if n[5] >= 9:
					n[6] += 5000
					n[4], n[5] = -1, -1,
					code.view_controller.instruction(None, n)
				#update wait time:
				n[6] += code.instructions.costs[operations[n[3]][n[4]][n[5]]]
				#generate bullets or shields:
				if make and len(make) == 6:
					bullets.append(make)
				if make and len(make) == 3:
					shields.append(make)
			
	#update bullets:
	for b in bullets[:]:
		tmp = dt
		while tmp>= 0:
			b[4], tmp = b[4] - dt, tmp - b[4]
			if b[4] <= 0:
				#move it forward:
				b[0] = (b[0] + b[2][0])%9
				b[1] = (b[1] + b[2][1])%9 #move it forward
				b[3] -= 1
				b[4] = code.instructions.bullet_time #reset timer
				if b[3] <= 0: #remove if dead.
					bullets.remove(b) 
					tmp = -1
			
	#update shields and remove dead:
	for s in shields[:]:
		s[2] -= dt
		if s[2] <= 0:
			shields.remove(s)
	
	#check collisions
	#bullet/shield:
	for b in bullets[:]:
		for s in shields:
			if b[0] == s[0] and b[1] == s[1]:
				bullets.remove(b)
				
	#bullet/nanite:
	for b in bullets[:]:
		for n in nanites:
			if b[0] == n[0] and b[1] == n[1]:
				if n[3] != b[5]: #if not same team
					n[3] = b[5] #change team
					n[4] = 0 #reset function
					n[5] = 0 #reset instruction
		
	code.view_controller.draw(game)
	
	#check for winner:
	winner = True
	team = nanites[0][3]
	for n in nanites:
		if n[3] != team:
			winner = False
	if winner:
		code.view_controller.game_over(team)
		running = False
		
	dt = clock.tick(60) * speed
	if dt > max_dt * speed: dt = max_dt * speed
